from dcim.models import Device, DeviceRole, DeviceType, Site
from ipam.models import Prefix
from dcim.choices import DeviceStatusChoices
from rest_framework import serializers
import json
import logging
from core.models import Job
from .models import AWXInventory

logger = logging.getLogger(__name__)

group_prefixes = {Site: "site_", DeviceRole: "devicerole_", DeviceType: "devicetype_", Prefix: "prefix_"}


class SiteSerializer(serializers.BaseSerializer):
    def to_representation(self, instance):
        return {
            "name": "{}{}".format(group_prefixes[Site], instance.slug.replace("-", "_")),
            "variables": json.dumps(
                {
                    "netbox_site_status": instance.status,
                }
            ),
        }


class DeviceRoleSerializer(serializers.BaseSerializer):
    def to_representation(self, instance):
        return {
            "name": "{}{}".format(group_prefixes[DeviceRole], instance.slug.replace("-", "_")),
            "variables": json.dumps({})
        }


class DeviceTypeSerializer(serializers.BaseSerializer):
    def to_representation(self, instance):
        return {
            "name": "{}{}".format(group_prefixes[DeviceType], instance.slug.replace("-", "_")),
            "description": instance.description,
            "variables": json.dumps(
                {
                    "netbox_devicetype_model": instance.model,
                }
            ),
        }


class PrefixSerializer(serializers.BaseSerializer):
    def to_representation(self, instance):
        return {
            "name": "{}{}".format(group_prefixes[Prefix], str(instance.prefix).replace(".", "_").replace("/", "_")),
            "variables": json.dumps({"netbox_prefix": "{}".format(str(instance.prefix))}),
        }


class InterfaceSerializer(serializers.BaseSerializer):
    def to_representation(self, instance):
        ip_addresses = []
        for ip in instance.ip_addresses.all():
            serializer = IPAddressSerializer(ip)
            ip_addresses.append(serializer.data)
        return {
            "name": instance.name,
            "mac": instance.mac_address,
            "ip_addresses": ip_addresses
        }


class IPAddressSerializer(serializers.BaseSerializer):
    def to_representation(self, instance):
        return {
            "address": str(instance.address),
            "dns_name": instance.dns_name
        }


class DeviceSerializer(serializers.BaseSerializer):
    def to_representation(self, instance):
        variables = {}
        variables["netbox_interfaces"] = []
        for interface in instance.interfaces.all():
            serializer = InterfaceSerializer(interface)
            variables["netbox_interfaces"].append(serializer.data)
        return {
            "name": instance.primary_ip4.dns_name,
            "description": instance.description,
            "enabled": instance.status == DeviceStatusChoices.STATUS_ACTIVE,
            "variables": json.dumps(variables),
        }


class AWXHostSerializer(serializers.BaseSerializer):
    def to_internal_value(self, data):
        return {
            "name": data.get('name'),
            "description": data.get('description'),
            "enabled": data.get('enabled'),
            "variables": data.get('variables')
        }


class AWXGroupSerializer(serializers.BaseSerializer):
    def to_internal_value(self, data):
        return {
            "name": data.get('name'),
            "variables": data.get('variables')
        }


serializers = {
    Site: SiteSerializer,
    DeviceRole: DeviceRoleSerializer,
    DeviceType: DeviceTypeSerializer,
    Prefix: PrefixSerializer,
    Device: DeviceSerializer,
}


def sync_host(inventory, sender, instance):
    serializer = serializers[sender](instance)
    host = inventory.get_host(serializer.data["name"])
    if host is None:
        inventory.create_host(serializer.data)
        host = inventory.get_host(serializer.data["name"])
    else:
        host_serializer = AWXHostSerializer(data=host)
        if not host_serializer.is_valid() or host_serializer.validated_data != serializer.data:
            inventory.update_host(host["id"], serializer.data)

    sync_host_group_association(inventory, host, DeviceRole, instance.device_role)
    sync_host_group_association(inventory, host, DeviceType, instance.device_type)
    sync_host_group_association(inventory, host, Site, instance.site)

    # print(Prefix.objects.filter(prefix__net_contains_or_equals=device.primary_ip4.address))


def sync_host_group_association(inventory, host, object_type, instance):
    serializer = serializers[object_type](instance)
    group_name = serializer.data["name"]
    groups = [
        group
        for group in host["summary_fields"]["groups"]["results"]
        if group["name"].startswith(group_prefixes[object_type])
    ]
    for group in groups:
        if group["name"] != group_name:
            inventory.disassociate_host_group(host["id"], group["id"])
    if group_name not in [group["name"] for group in groups]:
        group = inventory.get_group(group_name)
        inventory.associate_host_group(host["id"], group["id"])


def delete_host(inventory, sender, instance):
    serializer = serializers[sender](instance)
    inventory.delete_host(serializer.data["name"])


def sync_group(inventory, sender, instance):
    serializer = serializers[sender](instance)
    group = inventory.get_group(serializer.data["name"])
    if group is None:
        inventory.create_group(serializer.data)
        group = inventory.get_group(serializer.data["name"])
    else:
        group_serializer = AWXGroupSerializer(data=group)
        if not group_serializer.is_valid() or group_serializer.validated_data != serializer.data:
            print("Group will be updated due to:")
            print(group_serializer.validated_data)
            print(serializer.data)
            inventory.update_group(serializer.data["name"], serializer.data)


def delete_group(inventory, sender, instance):
    serializer = serializers[sender](instance)
    inventory.delete_group(serializer.data["name"])


def sync_all(job):
    inventory = job.object
    logger.info("Performing full inventory sync for inventory {}".format(inventory.inventory_id))
    job.start()
    for site in Site.objects.all():
        sync_group(inventory, Site, site)
    for device_role in DeviceRole.objects.all():
        sync_group(inventory, DeviceRole, device_role)
    for device_type in DeviceType.objects.all():
        sync_group(inventory, DeviceType, device_type)
    for device in Device.objects.all():
        if not device.primary_ip4 is None and device.primary_ip4.dns_name:
            sync_host(inventory, Device, device)
    job.terminate()
